"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = require("util");
const path = require("path");
const async = require("async");
const inquirer = require("inquirer");
const ora = require("ora");
const boxen = require("boxen");
const chalk = require("chalk");
const config_1 = require("../config");
const sonarwhal_1 = require("../sonarwhal");
const types_1 = require("../types");
const debug_1 = require("../utils/debug");
const get_as_uri_1 = require("../utils/get-as-uri");
const logger = require("../utils/logging");
const misc_1 = require("../utils/misc");
const resourceLoader = require("../utils/resource-loader");
const npm_1 = require("../utils/npm");
const each = util_1.promisify(async.each);
const debug = debug_1.debug(__filename);
const askForConfirm = (question) => {
    debug(`Asking for confirmation`);
    const questions = [{
            message: question,
            name: 'confirm',
            type: 'confirm'
        }];
    return inquirer.prompt(questions);
};
const askUserToCreateConfig = async () => {
    const question = `Do you want to create a new configuration?`;
    const launchInit = await askForConfirm(question);
    if (!launchInit.confirm) {
        return false;
    }
    const { default: initSonarwhalrc } = await Promise.resolve().then(() => require('./wizards/init'));
    const sonarwhalrcCreated = await initSonarwhalrc();
    if (!sonarwhalrcCreated) {
        return false;
    }
    logger.log(`Configuration file .sonarwhalrc was created.`);
    return true;
};
const askUserToUseDefaultConfiguration = async () => {
    const question = `A valid configuration file can't be found. Do you want to use the default configuration? To know more about the default configuration see: https://sonarwhal.com/docs/user-guide/#default-configuration`;
    const confirmation = await askForConfirm(question);
    return confirmation.confirm;
};
const showMissingAndIncompatiblePackages = (resources) => {
    if (resources.missing.length > 0) {
        logger.log(`The following ${resources.missing.length === 1 ? 'package is' : 'packages are'} missing:
    ${resources.missing.join(', ')}`);
    }
    if (resources.incompatible.length > 0) {
        logger.log(`The following ${resources.incompatible.length === 1 ? 'package is' : 'packages are'} incompatible:
    ${resources.incompatible.join(', ')}`);
    }
};
const askUserToInstallDependencies = async (resources) => {
    showMissingAndIncompatiblePackages(resources);
    const dependencies = resources.incompatible.concat(resources.missing);
    const question = [{
            message: `There ${dependencies.length === 1 ? 'is a package' : 'are packages'} from your .sonarwhalrc file not installed or with an incompatible version. Do you want us to try to install/update them?`,
            name: 'confirm',
            type: 'confirm'
        }];
    const answer = await inquirer.prompt(question);
    return answer.confirm;
};
const showDefaultMessage = () => {
    const defaultMessage = `${chalk.default.yellow(`Couldn't find any valid configuration`)}

Running sonarwhal with the default configuration.

Learn more about how to create your own configuration at:

${chalk.default.green('https://sonarwhal.com/docs/user-guide/')}`;
    logger.log(boxen(defaultMessage, {
        align: 'center',
        margin: 1,
        padding: 1
    }));
};
const getDefaultConfiguration = () => {
    showDefaultMessage();
    return { extends: ['web-recommended'] };
};
const getUserConfig = (actions) => {
    const configPath = (actions && actions.config) || config_1.SonarwhalConfig.getFilenameForDirectory(process.cwd());
    if (!configPath) {
        return getDefaultConfiguration();
    }
    debug(`Loading configuration file from ${configPath}.`);
    try {
        const resolvedPath = path.resolve(process.cwd(), configPath);
        const config = config_1.SonarwhalConfig.loadConfigFile(resolvedPath);
        return config || getDefaultConfiguration();
    }
    catch (e) {
        logger.error(e);
        return null;
    }
};
const messages = {
    'fetch::end': '%url% downloaded',
    'fetch::start': 'Downloading %url%',
    'scan::end': 'Finishing...',
    'scan::start': 'Analyzing %url%',
    'traverse::down': 'Traversing the DOM',
    'traverse::end': 'Traversing finished',
    'traverse::start': 'Traversing the DOM',
    'traverse::up': 'Traversing the DOM'
};
const getEvent = (event) => {
    if (event.startsWith('fetch::end')) {
        return 'fetch::end';
    }
    return event;
};
const setUpUserFeedback = (sonarwhalInstance, spinner) => {
    sonarwhalInstance.prependAny((event, value) => {
        const message = messages[getEvent(event)];
        if (!message) {
            return;
        }
        spinner.text = message.replace('%url%', misc_1.cutString(value.resource));
    });
};
const getDefaultOrCreateConfig = async (actions) => {
    const useDefault = await askUserToUseDefaultConfiguration();
    let userConfig;
    if (useDefault) {
        showDefaultMessage();
        userConfig = { extends: ['web-recommended'] };
    }
    else {
        const created = await askUserToCreateConfig();
        if (created) {
            userConfig = await getUserConfig();
        }
        else {
            logger.error(`Unable to find a valid configuration file. Please create a valid .sonarwhalrc file using 'sonarwhal --init'. `);
            return null;
        }
    }
    return config_1.SonarwhalConfig.fromConfig(userConfig, actions);
};
const getSonarwhalConfiguration = async (userConfig, actions) => {
    if (!userConfig) {
        return getDefaultOrCreateConfig(actions);
    }
    let config;
    try {
        config = config_1.SonarwhalConfig.fromConfig(userConfig, actions);
    }
    catch (err) {
        logger.error(err.message);
        config = await getDefaultOrCreateConfig(actions);
    }
    return config;
};
exports.sonarwhal = null;
exports.default = async (actions) => {
    const targets = get_as_uri_1.getAsUris(actions._);
    if (targets.length === 0) {
        return false;
    }
    const userConfig = await getUserConfig(actions);
    const config = await getSonarwhalConfiguration(userConfig, actions);
    if (!config) {
        return false;
    }
    let resources = resourceLoader.loadResources(config);
    if (resources.missing.length > 0 || resources.incompatible.length > 0) {
        const missingPackages = resources.missing.map((name) => {
            return `@sonarwhal/${name}`;
        });
        const incompatiblePackages = resources.incompatible.map((name) => {
            return `@sonarwhal/${name}@latest`;
        });
        if (!(await askUserToInstallDependencies(resources) &&
            await npm_1.installPackages(missingPackages) &&
            await npm_1.installPackages(incompatiblePackages))) {
            return false;
        }
        resources = resourceLoader.loadResources(config);
    }
    const invalidConfigRules = config_1.SonarwhalConfig.validateRulesConfig(config).invalid;
    if (invalidConfigRules.length > 0) {
        logger.error(`Invalid rule configuration in .sonarwhalrc: ${invalidConfigRules.join(', ')}.`);
        return false;
    }
    exports.sonarwhal = new sonarwhal_1.Sonarwhal(config, resources);
    const start = Date.now();
    const spinner = ora({ spinner: 'line' });
    let exitCode = 0;
    if (!actions.debug) {
        spinner.start();
        setUpUserFeedback(exports.sonarwhal, spinner);
    }
    const endSpinner = (method) => {
        if (!actions.debug && spinner[method]) {
            spinner[method]();
        }
    };
    const hasError = (reports) => {
        return reports.some((result) => {
            return result.severity === types_1.Severity.error;
        });
    };
    const print = async (reports, target) => {
        if (hasError(reports)) {
            endSpinner('fail');
        }
        else {
            endSpinner('succeed');
        }
        await each(exports.sonarwhal.formatters, async (formatter) => {
            await formatter.format(reports, target);
        });
    };
    exports.sonarwhal.on('print', print);
    for (const target of targets) {
        try {
            const results = await exports.sonarwhal.executeOn(target);
            if (hasError(results)) {
                exitCode = 1;
            }
            await print(results, target.href);
        }
        catch (e) {
            exitCode = 1;
            endSpinner('fail');
            debug(`Failed to analyze: ${target.href}`);
            debug(e);
        }
    }
    await exports.sonarwhal.close();
    debug(`Total runtime: ${Date.now() - start}ms`);
    return exitCode === 0;
};
